# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

from pyrddl.parser import RDDLParser
from tfrddlsim.rddl2tf.compiler import Compiler

from tfmdp.train.policy import DeepReactivePolicy

import numpy as np
import tensorflow as tf

import unittest


class TestDeepReactivePolicy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # hyper-parameters
        cls.batch_size = 64
        cls.horizon = 15

        # RDDL
        with open('rddl/Reservoir.rddl', mode='r') as file:
            RESERVOIR = file.read()

        # parser
        parser = RDDLParser()
        parser.build()
        rddl = parser.parse(RESERVOIR)

        # compiler
        cls.compiler = Compiler(rddl, batch_mode=True)
        cls.initial_state = cls.compiler.compile_initial_state(cls.batch_size)

        # Deep Reactive Policy
        cls.channels = 8
        cls.layers = [128, 64]
        cls.policy = DeepReactivePolicy(cls.compiler, cls.channels, cls.layers)
        cls.action = cls.policy(cls.initial_state, cls.horizon)

    def test_state_inputs(self):
        self.assertIsInstance(self.policy.state_inputs, tuple)
        self.assertEqual(len(self.policy.state_inputs), len(self.initial_state))
        for layer, fluent in zip(self.policy.state_inputs, self.initial_state):
            self.assertIsInstance(layer, tf.Tensor)
            self.assertEqual(layer.dtype, fluent.dtype)
            self.assertEqual(len(layer.shape), 2)
            self.assertEqual(layer.shape.as_list()[0], self.batch_size)
            self.assertEqual(np.prod(layer.shape.as_list()), np.prod(fluent.shape.as_list()))

    def test_input_layer(self):
        self.assertIsInstance(self.policy.input_layer, tf.Tensor)

        sizes = list(np.prod(fluent.shape.as_list()) / self.batch_size for fluent in self.initial_state)
        total_size = sum(sizes)
        shape = [self.batch_size, self.channels * total_size]
        self.assertListEqual(self.policy.input_layer.shape.as_list(), shape)

        with self.compiler.graph.as_default():
            state_fluents = self.compiler.state_fluent_ordering
            state_size = self.compiler.state_size
            for name, shape in zip(state_fluents, state_size):
                name = name.replace('/', '-')
                size = np.prod(shape)
                units = self.channels * size

                kernel = 'layer0/{}/dense/kernel'.format(name)
                vars = tf.global_variables(kernel)
                self.assertEqual(len(vars), 1)
                self.assertListEqual(vars[0].shape.as_list(), [size, units])

                bias = 'layer0/{}/dense/bias'.format(name)
                vars = tf.global_variables(bias)
                self.assertEqual(len(vars), 1)
                self.assertListEqual(vars[0].shape.as_list(), [units,])

    def test_hidden_layers(self):
        self.assertIsInstance(self.policy.hidden, tuple)
        self.assertEqual(len(self.policy.hidden), len(self.layers))

        with self.compiler.graph.as_default():
            for l, units in enumerate(self.layers):
                vars = tf.global_variables('hidden{}'.format(l+1))
                self.assertEqual(len(vars), 2)

                kernel = 'hidden{}/dense/kernel'.format(l+1)
                vars = tf.global_variables(kernel)
                self.assertEqual(len(vars), 1)
                self.assertEqual(vars[0].shape[1], units)

                bias = 'hidden{}/dense/bias'.format(l+1)
                vars = tf.global_variables(bias)
                self.assertEqual(len(vars), 1)
                self.assertEqual(vars[0].shape[0], units)

    def test_output_layer(self):
        action_fluents = self.compiler.action_fluent_ordering
        action_size = self.compiler.action_size
        self.assertIsInstance(self.policy.output_layer, tuple)
        self.assertEqual(len(self.policy.output_layer), len(action_fluents))

        with self.compiler.graph.as_default():

            for name, shape, layer in zip(action_fluents, action_size, self.policy.output_layer):
                self.assertIsInstance(layer, tf.Tensor)
                self.assertEqual(layer.dtype, tf.float32)
                self.assertEqual(len(layer.shape), 2)
                self.assertEqual(layer.shape[0], self.batch_size)
                self.assertEqual(layer.shape[1], np.prod(shape))

                name = name.replace('/', '-')
                vars = tf.global_variables('output/{}'.format(name))
                self.assertEqual(len(vars), 2)

                kernel = 'output/{}/dense/kernel'.format(name)
                vars = tf.global_variables(kernel)
                self.assertEqual(len(vars), 1)

                bias = 'output/{}/dense/bias'.format(name)
                vars = tf.global_variables(bias)
                self.assertEqual(len(vars), 1)

    def test_action_outputs(self):
        action_fluents = self.compiler.action_fluent_ordering
        action_size = self.compiler.action_size
        action_dtype = self.compiler.action_dtype
        self.assertIsInstance(self.policy.action_outputs, tuple)
        self.assertEqual(len(self.policy.action_outputs), len(action_fluents))
        for name, shape, dtype, tensor in zip(action_fluents, action_size, action_dtype, self.policy.action_outputs):
            self.assertIsInstance(tensor, tf.Tensor)
            self.assertEqual(tensor.dtype, dtype)
            self.assertListEqual(tensor.shape.as_list(), [self.batch_size] + list(shape))



# This file is part of tf-mdp.

# tf-mdp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-mdp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-mdp. If not, see <http://www.gnu.org/licenses/>.

from pyrddl.parser import RDDLParser
from tfrddlsim.rddl2tf.compiler import Compiler

from tfmdp.train.policy import DeepReactivePolicy
from tfmdp.train.optimizer import PolicyOptimizer

import numpy as np
import tensorflow as tf

import unittest


class TestPolicyOptimizer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # hyper-parameters
        cls.learning_rate = 0.0001
        cls.epochs = 32
        cls.batch_size = 64
        cls.horizon = 20

        # RDDL
        with open('rddl/Reservoir.rddl', mode='r') as file:
            RESERVOIR = file.read()

        # parser
        parser = RDDLParser()
        parser.build()
        rddl = parser.parse(RESERVOIR)

        # compiler
        cls.compiler = Compiler(rddl, batch_mode=True)
        cls.initial_state = cls.compiler.compile_initial_state(cls.batch_size)

        # policy
        cls.channels = 4
        cls.layers = [64, 32, 16]
        cls.policy = DeepReactivePolicy(cls.compiler, cls.channels, cls.layers)

        # optimizer
        cls.optimizer = PolicyOptimizer(cls.compiler, cls.policy)
        cls.optimizer.build(cls.learning_rate, cls.batch_size, cls.horizon)

    def test_state_trajectory(self):
        states = self.optimizer.states
        state_size = self.compiler.state_size
        self.assertIsInstance(states, tuple, 'state trajectory is factored')
        self.assertEqual(len(states), len(state_size), 'state trajectory has all states fluents')
        for fluent, fluent_size in zip(states, state_size):
            tensor_size = [self.batch_size, self.horizon] + list(fluent_size)
            self.assertIsInstance(fluent, tf.Tensor, 'state fluent is a tensor')
            self.assertListEqual(fluent.shape.as_list(), tensor_size, 'fluent size is [batch_size, horizon, state_fluent_size]')

    def test_action_trajectory(self):
        actions = self.optimizer.actions
        action_size = self.compiler.action_size
        self.assertIsInstance(actions, tuple, 'action trajectory is factored')
        self.assertEqual(len(actions), len(action_size),
            'action trajectory has all actions fluents')
        for fluent, fluent_size in zip(actions, action_size):
            tensor_size = [self.batch_size, self.horizon] + list(fluent_size)
            self.assertIsInstance(fluent, tf.Tensor, 'action fluent is a tensor')
            self.assertListEqual(fluent.shape.as_list(), tensor_size,
                'fluent size is [batch_size, horizon, action_fluent_size]')

    def test_rewards_trajectory(self):
        rewards = self.optimizer.rewards
        rewards_shape = [self.batch_size, self.horizon, 1]
        self.assertIsInstance(rewards, tf.Tensor, 'reward trajectory is a tensor')
        self.assertListEqual(rewards.shape.as_list(), rewards_shape ,
            'reward shape is [batch_size, horizon, 1]')

    def test_total_reward(self):
        total_reward = self.optimizer.total_reward
        self.assertIsInstance(total_reward, tf.Tensor, 'total reward is a tensor')
        self.assertEqual(total_reward.dtype, tf.float32, 'total reward is a real tensor')
        self.assertListEqual(total_reward.shape.as_list(), [self.batch_size],
            'total reward has a scalar value for each trajectory')

    def test_avg_total_reward(self):
        avg_total_reward = self.optimizer.avg_total_reward
        self.assertIsInstance(avg_total_reward, tf.Tensor, 'average total reward is a tensor')
        self.assertEqual(avg_total_reward.dtype, tf.float32, 'average total reward is a real tensor')
        self.assertListEqual(avg_total_reward.shape.as_list(), [], 'average total reward is a scalar')

    def test_optimization_objective(self):
        loss = self.optimizer.loss
        self.assertIsInstance(loss, tf.Tensor, 'loss function is a tensor')
        self.assertEqual(loss.dtype, tf.float32, 'loss function is a real tensor')
        self.assertListEqual(loss.shape.as_list(), [], 'loss function is a scalar')

    def test_loss_optimizer(self):
        optimizer = self.optimizer._optimizer
        self.assertIsInstance(optimizer, tf.train.RMSPropOptimizer)
        train_op = self.optimizer._train_op
        self.assertEqual(train_op.name, 'policy_optimizer/RMSProp')

    def test_run(self):
        self.optimizer.run(self.epochs, show_progress=False)
